/***************************************************************************
 *   Copyright (C) 2013 by hovercraft:                                     *
 *		https://github.com/hovercraft-github                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** @file *********************************************************************
 *
 * @brief Circular buffer.
 *
 * Provides double-ended queue stored in user-provided static circular buffer
 * for placing arbitrary objects.
 * Collections of up to 126 elements supported.
 * Attention: this implementation differs from the standard deque in that
 * an access to stored elements is realized via pointers to the circular buffer location
 * rather than over reference/object copies.
 * In fact, pointer returned by pop_back, pop_front methods refer
 * to the already released object in the queue.
 * Due to static nature of the circular buffer, it's o'k to use such a pointers,
 * but only until some push_back, push_front methods called. This should be taken into account
 * especially while writing interrupt routings.
 *****************************************************************************/

#ifndef CIRCUMGIR_H_INCLUDED
#define CIRCUMGIR_H_INCLUDED


/** @class
 * @brief Base abstract class for collections.
 *
*/
template <class A_CollectionItemType> class TAbstractCollector
{
  public:
    TAbstractCollector() {};
    virtual A_CollectionItemType * get(char ix) = 0;
    virtual char push_front(A_CollectionItemType *src) = 0;
    virtual A_CollectionItemType * pop_back() = 0;
};

/** @class
 * @brief Base template class for collection items.
 * Provides operator new for in-place allocating object inside the collector circular buffer.
*/
template <class A_CollectionItemType> class TCollectionItem
{
  public:
    void* operator new(size_t size, TAbstractCollector<A_CollectionItemType> &collector) {return collector.get(collector.push_front((A_CollectionItemType *)0));};
    void operator delete(void*) {};
};

/** @class
 * @brief Base template class for d-queues of arbitrary objects.
 * Provides general double-ended queue functionality.
 * Inherits both from TCollectionItem and TAbstractCollector so
 * queue of queues can be created.
*/
template <class A_Type> class TCircumgyration : public TCollectionItem< TCircumgyration<A_Type> >, public TAbstractCollector<A_Type>
{
  //private:
  //  friend class GXXBunch;
  protected:


    /**
        Pointer to the user-provided buffer.
    */
    A_Type *m_buff;
    /**
        Size of buffer in number of elements.
    */
    char m_buff_size;
    /**
        Index of the head element in the buffer.
    */
    char m_head;
    /**
        Index of the tail element in the buffer
    */
    char m_tail;
    /**
        Get the insertion point before a head for the next element.
        @return the first vacant position index
    */
    char vacant();


  public:
    //void* operator new(size_t size, TCircumgyration<A_Type> &target) {return &target;};
    //void operator delete(void*) {};

    /**
     * @brief in-place constructor for TCircumgyration class
     * @param[in] buffer: user-provided circular buffer
     * @param[in] size:   buffer size in number of elements
    */
    TCircumgyration(A_Type *buffer, char size);
    /**
     * @brief in-place constructor for TCircumgyration class.
     * Initializes circular buffer head and tail to the specified values.
     * @param[in] buffer: user-provided circular buffer
     * @param[in] size:   buffer size in number of elements
     * @param[in] head:   head element index
     * @param[in] tail:   tail element index
    */
    TCircumgyration(A_Type *buffer, char size, char head, char tail);

    /** put: put the object into collection at the specified index
    @param[in] A_Type *: object pointer
    @param[in] char: index of position in the collection where to place object
    */
    void put(A_Type *src, char ix);
    /** get: get pointer to the element at specified index
    Failed boundary check on index results in NULL pointer return.
    @param[in] char ix: index of the element
    */
    virtual A_Type * get(char ix) {if (ix < 0) return (A_Type *)0; return m_buff+ix;};


    /** @name INDEX_FUNCTIONS
        Collection element index iteration functions.
        Returns index of the element corresponding to the function name.
    */
    ///@{
    char head() {return m_head;};
    char tail() {return m_tail;};
    char first() {if (m_tail < 0) return -1; return m_tail;};
    char next(char ix);
    char prev(char ix);
    char last() {if (m_tail < 0) return -1; return m_head-1;};
    ///@}

    /** room: how many elements could be placed next into collection
    @return number of free elements in the circular buffer
    */
    char room();
    /** count: number of elements
    @return number of elements in the circular buffer
    */
    char count() {return m_buff_size-room();};


    /** @name DQUEUE_METHODS
    Standard d-queue interface methods:
    */
    ///@{
    /** max_size: maximum number of elements we can insert into collection.
    @return maximum number of elements
    */
    char max_size() {return m_buff_size-1;};
    /** size: number of elements
    Just for convenience - synonym to count()
    @return the number of elements in the deque container.
    */
    char size() {return count();};
    /** push_front: Insert element at beginning.
    The start of the inserted element corresponds to the head position in the circular buffer.
    @param[in] src: pointer of the source element
    @return index of the element in the circular buffer
    */
    virtual char push_front(A_Type *src);
    /** front: Access the first element without removing it.
    The start of the first element corresponds to the head position in the circular buffer.
    @return pointer to the element
    */
    virtual A_Type * front() {return get(m_head);};
    /** back: Access the last element without removing it.
    The start of the last element corresponds to the tail position in the circular buffer.
    @return pointer to the element
    */
    virtual A_Type * back() {return get(m_tail);}; //virtual A_Type * get() {return get(m_tail);};
    /** pop_front: Access the first element in container.
    !! Attention: when this function returns, the space of the element is already marked free for the collector,
    so in the multi-thread environment measures must be taken to prevent collection updating
    until we finish using extracted element !!
    @return pointer to the element
    */
    virtual A_Type * pop_front();
    /** pop_back: Access the last element in container.
    !! Attention: when this function returns, the space of the element is already marked free for the collector,
    so in the multi-thread environment measures must be taken to prevent collection updating
    until we finish using extracted element !!
    @return pointer to the element
    */
    virtual A_Type * pop_back();
    ///@}

    /** cutoff_back: remove last element
    Release space of the last element (the snake tail) in the circular buffer.
    @return the number of elements left after operation.
    */
    virtual char cutoff_back() {pop_back(); return count();};
    /** cutoff_front: remove first element
    Release space of the first element (the snake head) in the circular buffer.
    @return the number of elements left after operation.
    */
    virtual char cutoff_front() {pop_front(); return count();};
};

/** @class TGirString
 * @brief Simple character d-queue class.
 * Stores character queue in the circular buffer.
 * Inherits from the TCircumgyration class.
 *
*/
class TGirString : public TCircumgyration<char>
{
  public:
  TGirString(char *buffer, int size) : TCircumgyration<char>(buffer, size) {};
};

/**
 * @brief in-place constructor for TCircumgyration class
 * @param buffer: user-provided circular buffer
 * @param size:   buffer size in number of elements
*/
template <class A_Type> TCircumgyration<A_Type>::TCircumgyration(A_Type *buffer, char size)
{
  m_buff = buffer;
  m_buff_size = size; // number of elements = sizeof(m_buff)/sizeof(A_Type);
  m_head = m_tail = -1;
}

/**
 * @brief in-place constructor for TCircumgyration class.
 * Initializes circular buffer head and tail to the specified values.
 * @param buffer: user-provided circular buffer
 * @param size:   buffer size in number of elements
 * @param head:   head element index
 * @param tail:   tail element index
*/
template <class A_Type> TCircumgyration<A_Type>::TCircumgyration(A_Type *buffer, char size, char head, char tail)
{
  m_buff = buffer;
  m_buff_size = size; // number of elements = sizeof(m_buff)/sizeof(A_Type);
  m_head = head;
  m_tail = tail;
}

template <class A_Type> void TCircumgyration<A_Type>::put(A_Type *src, char ix)
{
  if ((ix < 0) || (ix >= m_buff_size))
    return;
  unsigned char *dst = (unsigned char *)(m_buff + ix);
  if (src)
    {
    for (char i = 0; i < sizeof(A_Type); i++)
      *(dst + i) = *(((unsigned char *)src) + i);
    }
   else
    {
    for (char i = 0; i < sizeof(A_Type); i++)
      *(dst + i) = (unsigned char)0;
    }
}

template <class A_Type> char TCircumgyration<A_Type>::vacant()
{
  if (m_tail < 0)
    return 0;
  char next = m_head + 1;
  if (next >= m_buff_size)
    next = 0;
  if (next == m_tail)
    return -1; //Buffer overflow
  return next;
}

template <class A_Type> A_Type * TCircumgyration<A_Type>::pop_front()
{
    A_Type * ret;
    if (m_head < 0)
        return (A_Type *)0;
    ret = m_buff + m_head;
    char cnt = count();
    if (cnt == 1)
        {
        m_head = m_tail = -1; //No more elements in this queue
        }
    else
        {
        m_head -= 1;
        if (m_head < 0)
            m_head = m_buff_size - 1;
        }
    return ret;
}

template <class A_Type> char TCircumgyration<A_Type>::push_front(A_Type *src)
{
  char ix = vacant();
  if (ix >= 0)
    {
    put(src, ix);
    m_head = ix;
    if (m_tail < 0)
      m_tail = m_head;
    };
  return ix;
}

template <class A_Type> A_Type * TCircumgyration<A_Type>::pop_back()
{
  A_Type * ret;
  if (m_tail < 0)
    return (A_Type *)0;
  char ix = next(m_tail);
  ret = m_buff + m_tail;
  m_tail = ix;
  if (m_tail < 0)
    m_head = m_tail;
  return ret;
}

template <class A_Type> char TCircumgyration<A_Type>::room()
{
  char room;
  if (m_tail < 0)
    return m_buff_size;
  else if (m_tail <= m_head)
    room = m_buff_size - m_head + m_tail - 1;
  else
    room = m_tail - m_head - 1;
  return room;
}

template <class A_Type> char TCircumgyration<A_Type>::next(char ix)
{
  if (m_tail < 0)
       return -1; //Buffer empty
  char next = ix + 1;
  if (next >= m_buff_size)
    next = 0;
  if (m_tail > m_head)
    {
     if ((next <= m_tail) && (next > m_head))
       return -1; //Out of buffer
    }
  else
     if ((next <= m_tail) || (next > m_head))
       return -1; //Out of buffer
  return next;
}

template <class A_Type> char TCircumgyration<A_Type>::prev(char ix)
{
  if (m_tail < 0)
       return -1; //Buffer empty
  char prev = ix - 1;
  if (prev < 0)
    prev = m_buff_size - 1;
  if (prev < m_tail)
      {
    if (m_tail < m_head)
      {
         return -1; //Out of buffer
      }
    else
      {
       if (prev >= m_head)
         return -1; //Out of buffer
      };
      };
  return prev;
}

#endif // CIRCUMGIR_H_INCLUDED
